/* ************************************************************************** */
/*                                       _____           _                    */
/*  game.c                              |_   _|__  _   _| |__   ___ _   _     */
/*  | Project : TuQuoque                  | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/08/08 15:03:04                               |___/     */
/*                                                                            */
/* ************************************************************************** */

#include <stdlib.h>
#include <string.h>
#include <keyboard.h>
#include <display.h>
#include <main.h>

/* player should have more life than enemies */
#define default_player_life 100
#define default_ai_life 10
/* max number of enemies spawn, more would be useless */
#define enemies_max 256
/* other constants useful to manipulate characters tab */
#define characters_len (1 + enemies_max)
#define enemies_offset 1
/* lephe said more would include fake presses */
#define keybuf_size 3
/* font info */
#define font_char_w 5
#define font_char_h 7

/**
 *	characters:
 *	The characters tab. Includes player and enemies.
 */

static character_t *characters; /* allocated dynamically -- see `game` */
static character_t *player; /* adjusted when allocated */
static character_t *enemies; /* adjusted when allocated */

/**
 *	ais:
 *	Artificial intelligence data.
 */

static ai_t *ais; /* allocated dynamically -- see `game` */

/**
 *	dcharacters:
 *	Display all characters.
 *
 *	All character follow the same model, so steps and arms are the same.
 *	The player is required because the camera is centered on him.
 *
 *	Again, a lot of experimental values because I didn't want them in the
 *	character images (they'd take up too much space).
 */

static void dcharacters(void)
{
	for (int i = 0; i < 1 + enemies_max; i++) {
		character_t *character = &characters[i];
		if (character->life <= 0)
			continue;

		int x = DISPLAY_WIDTH / 2 - 8 + (character->x - player->x);
		int y = DISPLAY_HEIGHT / 2 - 8 + (character->y - player->y);

		/* quickly check if we are in bounds */
		if (x < -15 || x >= DISPLAY_WIDTH || y < -15 || y >= DISPLAY_HEIGHT)
			continue;

		/* then draw the character base */
		dimage_part(x, y, character->image,
			16 * character->direction, 0, 16, 16);

		/* then draw feet */
		character->step %= 32;
		int step = character->step / 8;
		dimage_part(x, y, &img_feet, character->direction * 16,
			16 * (step % 2 ? step - (step > 2) : 0), 16, 16);

		/* draw weapon */
		if (character->weapon_status != wps_hidden) {
			/* adjust weapon step */
			character->weapon_step %= 2;
			/* get weapon image */
			Image *wp_image;
			if (character->weapon == wp_hand) wp_image = &img_arms;
			else wp_image = &img_knives;
			/* put */
			dimage_part(x, y, wp_image,
				16 * character->direction, 16 * character->weapon_step,
				16, 16);
		}
	}
}

/**
 *	dmap:
 *	Display map from player point of view.
 *
 *	Only displays the tiles that are potentially on screen.
 *	(yeah, cause border tiles can not appear in some extreme cases,
 *	but nevermind, it's not an extreme loss to draw them anyway)
 */

static void dmap(void)
{
	/* get big and micro player positions */
	const int pgx = player->x / 16, pgy = player->y / 16;
	const int pmx = player->x % 16, pmy = player->y % 16;
	/* get map dimensions and bytes */
	const int mw = map->w, mh = map->h;
	const uint8_t *mb = map->b;

	/* display loop */
	for (int y = -3; y < 4; y++) {
		for (int x = -5; x < 6; x++) {
			/* get tile */
			int tile = 1; /* default tile is full tile for all maps */
			int tx = pgx + x, ty = pgy + y;
			if (tx >= 0 && tx < mw && ty >= 0 && ty < mh)
				tile = mb[ty * mw + tx];
			/* display tile */
			if (tile--) {
				dimage_part(
					DISPLAY_WIDTH  / 2 - 8 + x * 16 - pmx,
					DISPLAY_HEIGHT / 2 - 8 + y * 16 - pmy,
					&img_tiles,
					tile * 16, 0, 16, 16);
			}
		}
	}
}

/**
 *	dinterface:
 *	Draw interface.
 *
 *	Draws things like life, scores, and so on.
 *	In fact, only player life for now.
 */

static void dinterface(void)
{
	Image *lifebar = &img_life;

	/* draw player life */
	/* - base image - */
	dclear_area(DISPLAY_WIDTH - lifebar->width,
		DISPLAY_HEIGHT - lifebar->height,
		DISPLAY_WIDTH, DISPLAY_HEIGHT);
	dimage(DISPLAY_WIDTH - lifebar->width,
		DISPLAY_HEIGHT - lifebar->height, lifebar);
	int x = DISPLAY_WIDTH - lifebar->width + 11;
	int y = DISPLAY_HEIGHT - lifebar->height + 2;
	int max = x + player->life * 50 / default_player_life;
	dreverse_area(x, y, max, y + 4);
}

/**
 *	ai_decide:
 *	Lets AI decide what they want to do.
 *
 *	No pathfinding algorithm because flemme
 */

static void ai_decide(void)
{
	int px = player->x, py = player->y;

	for (int i = 0; i < enemies_max; i++) {
		character_t *ai = &enemies[i];
		if (ai->life <= 0)
			continue;
		ai_t *brain = &ais[i];

		/* make proxies */
		int ax = ai->x, ay = ai->y;

		/* check if player is nearby */
		int nearby = 1;
		if (py >= ay - 17 && py <= ay - 15
		 && px >= ax - 8 && px <= ax + 8)
			ai->direction = dir_up;
		else if (py >= ay + 15 && py <= ay + 17
		 && px >= ax - 8 && px <= ax + 8)
			ai->direction = dir_down;
		else if (px >= ax - 17 && px <= ax - 15
		 && py >= ay - 8 && py <= ay + 8)
			ai->direction = dir_left;
		else if (px >= ax + 15 && px <= ax + 17
		 && py >= ay - 8 && py <= ay + 8)
			ai->direction = dir_right;
		else
			nearby = 0;

		/* if nearby, attack */
		if (nearby) {
			if (!brain->ticks_before_changing_wps--) {
				ai->weapon_status = !ai->weapon_status;
				brain->ticks_before_changing_wps = 5;
			}
			continue;
		}

		/* reset ticks before changing wps */
		ai->weapon_status = wps_hidden;
		brain->ticks_before_changing_wps = 0;

		/* then move towards player */
		ai->vx = 0, ai->vy = 0;
		if (ax < px - 8)
			ai->vx = 1;
		else if (ax > px + 8)
			ai->vx = -1;
		if (ay < py - 8)
			ai->vy = 1;
		else if (ay > py + 8)
			ai->vy = -1;
	}
}

/**
 *	ai_spawn:
 *	Spawn AIs wave.
 */

static void ai_spawn(void)
{
	Image* ai_images[] = {&img_character_ia1, &img_character_ia2,
		&img_character_ia3};

	for (int i = 0; i < enemies_max; i++) {
		if (i >= map_ai->n)
			continue;

		/* initialize */
		enemies[i] = (character_t){
			.life = default_ai_life,
			.x = map_ai->spawns[i].x * 16,
			.y = map_ai->spawns[i].y * 16,
			.vx = 0, .vy = 0,
			.direction = dir_down, .step = 0,
			.weapon = wp_knife, .weapon_status = wps_trig,
			.image = ai_images[rand() % 2]
		};
		ais[i] = (ai_t){
			.ticks_before_changing_wps = 0
		};
	}
}

/**
 *	move_characters:
 *	Moves all characters -- takes collisions into account.
 *
 *	Take this as an example :
 *
 *	        character->x
 *	             |
 *	character->y-O.
 *	             ..
 *
 *	Here, (character->x / 16, character->y / 16) are the coordinates of the O
 *	square (a tile is 16*16px long), and (player.x % 16, player.y % 16) are
 *	the relative coordinates in pixels to these ones, or "how the player is
 *	relatively to this square". It goes to right and down.
 *
 *	As a consequence, when player.x % 16 == 0, the player doesn't go beyond
 *	the horizontal limit, and it goes the same for player.y.
 *
 *	So in each big condition, we have :
 *	- the player tries to go left and go outside the map ;
 *	- the player tries to go left and the left square is not air ;
 *	- the player tries to go left, overflows on the bottom square and the
 *		bottom-left square is not air ;
 *	Same goes for when the player wants to go right, and same goes for when
 *	the player wants to go upwards or downwards.
 *
 *	You might have understood it as well, but abscissa evolution comes before
 *	ordinate evolution, which means when the player can move on one of the two
 *	awes but not the two at a time, it will move on the x-axis.
 */

static void move_characters(void)
{
	for (int i = 0; i < characters_len; i++) {
		character_t *character = &characters[i];
		if (character->life <= 0)
			continue;

		/* get decision vectors to local */
		int vx = character->vx, vy = character->vy;

		/* first of all, check direction */
		if (vy) character->direction = (vy < 0) ? dir_up : dir_down;
		else if (vx) character->direction = (vx < 0) ? dir_left : dir_right;

		/* then check x-axis evolution */
		int px = character->x, py = character->y;
		int base = (py / 16) * map->w + px / 16;
		if (((vx < 0 && !(px % 16)) && (px <= 16
			|| map->b[base - 1]
			|| ((py % 16) && map->b[base + map->w - 1])))
		 || ((vx > 0) && (px >= map->w * 16 - 16
			|| map->b[base + 1]
			|| ((py % 16) && map->b[base + map->w + 1]))))
			vx = 0;

		/* check collision with other characters after horizontal movement */
		px = character->x + vx;
		for (int j = 0; j < 1 + enemies_max; j++) {
			character_t *other = &characters[j];
			if (j == i || other->life <= 0)
				continue;

			int ox = other->x, oy = other->y;
			if (py > oy - 16 && py < oy + 16
			 && px > ox - 16 && px < ox + 16)
				vx = 0;
		}

		/* then check y-axis evolution */
		base = (py / 16) * map->w + px / 16;
		if (((vy < 0 && !(py % 16)) && (py <= 16
			|| map->b[base - map->w]
			|| ((px % 16) && map->b[base + 1 - map->w])))
		 || ((vy > 0) && (py >= map->h * 16 - 16
			|| map->b[base + map->w]
			|| ((px % 16) && map->b[base + 1 + map->w]))))
			vy = 0;

		/* check collision with other characters after vertical movement */
		py = character->y + vy;
		for (int j = 0; j < 1 + enemies_max; j++) {
			character_t *other = &characters[j];
			if (j == i || other->life <= 0)
				continue;

			int ox = other->x, oy = other->y;
			if (py > oy - 16 && py < oy + 16
			 && px > ox - 16 && px < ox + 16)
				vy = 0;
		}

		/* update positions with checked vectors */
		character->x += vx;
		character->y += vy;
		character->cvx = vx;
		character->cvy = vy;
		if (vx || vy)
			character->step++;

		/* make damage */
		if (character->weapon_status == wps_trig) {
			/* get expected coordinates */
			int ex = character->x, ey = character->y;
			int hor;
			switch (character->direction) {
				case dir_up:    ey -= 16; hor = 0; break;
				case dir_down:  ey += 16; hor = 0; break;
				case dir_left:  ex -= 16; hor = 1; break;
				case dir_right: ex += 16; hor = 1; break;
				default:                  hor = 0; break;
			}
			/* check damage */
			int damage = 5; // TODO : with weapon ?
			/* check for each enemy */
			for (int j = 0; j < 1 + enemies_max; j++) {
				character_t *other = &characters[j];
				/* check if character is alive */
				/* we could check if i != j, but coordinates can't be */
				/* the one we expect. */
				if (other->life <= 0)
					continue;
				/* check if coordinates are expected */
				int ox = other->x, oy = other->y;
				if ((!hor && (ox < ex - 8 || ox > ex + 8 
					|| oy < ey - 1 || oy > ey + 1))
				 || ( hor && (oy < ey - 8 || oy > ey + 8
					|| ox < ex - 1 || ox > ex + 1)))
					continue;
				/* give damage !!! */
				other->life -= damage;
				break;
			}
		}

		/* check weapon status evolution */
		if (character->weapon_status == wps_trig)
			character->weapon_status = wps_out, character->weapon_step++;
	}
}

/**
 *	is_pressed_in:
 *	Check if a key is pressed in a pressed key buffer.
 *
 *	The current multigetkey implementation is quite annoying to use.
 *	But since we don't yet have SDL-like events, we use this function.
 *
 *	@arg	key		key to check
 *	@arg	buf		pressed key buffer
 *	@return			a boolean indicating if yes or no the key is pressed in buf
 */

static int is_pressed_in(int key, int *buf)
{
	int size = keybuf_size;
	while (size--) {
		if (*buf == key)
			return (1);
		else if (*buf++ == KEY_NONE)
			return (0);
	}
	return (0);
}

/**
 *	game:
 *	Game menu.
 *
 *	Plays game. Wowzy.
 *	Yeah, also, I know the key event detection is horrible, but as there are
 *	no key release or events in global yet in gint, I use the multigetkey
 *	method (which is the only method for now).
 *
 *	@arg	last	last menu
 */

void *game(void *last)
{
	/* initialize all characters (player and enemies) */
	/* - allocate - */
	characters = calloc(characters_len, sizeof(character_t));
	player = &characters[0];
	enemies = &characters[enemies_offset];
	/* - initialize all lives to 0 - */
	for (int i = 0; i < characters_len; i++)
		characters[i].life = 0;
	/* - initialize player - */
	*player = (character_t){
		.life = default_player_life,
		.x = map->spawn_x * 16,
		.y = map->spawn_y * 16,
		.vx = 0, .vy = 0,
		.direction = dir_down, .step = 0,
		.weapon = wp_hand, .weapon_status = wps_hidden,
		.image = &img_character_cesar
	};
	/* - allocate AIs - */
	ais = calloc(enemies_max, sizeof(ai_t));
	/* - initialize AIs - */
	ai_spawn();

	/* initialize key buffers */
	int keybuf_current[keybuf_size] = {KEY_NONE};
	int keybuf_last[keybuf_size] = {KEY_NONE};
	multigetkey(keybuf_current, keybuf_size, 0);

	/* game loop */
	while (1) {
		/* drawing */
		dclear();
		dmap();
		dcharacters();
		dinterface();
		dupdate();

		/* move character */
		move_characters();
		if (player->life <= 0)
			break;

		/* event gestion */
		memcpy(keybuf_last, keybuf_current, keybuf_size * sizeof(int));
		multigetkey(keybuf_current, keybuf_size, 1);

		/* - check if the player wants to move its character - */
		int old_up = is_pressed_in(KEY_UP, keybuf_last),
			new_up = is_pressed_in(KEY_UP, keybuf_current);
		player->vy -= (!old_up && new_up) - (old_up && !new_up);
		int old_dw = is_pressed_in(KEY_DOWN, keybuf_last),
			new_dw = is_pressed_in(KEY_DOWN, keybuf_current);
		player->vy += (!old_dw && new_dw) - (old_dw && !new_dw);
		int old_lf = is_pressed_in(KEY_LEFT, keybuf_last),
			new_lf = is_pressed_in(KEY_LEFT, keybuf_current);
		player->vx -= (!old_lf && new_lf) - (old_lf && !new_lf);
		int old_rg = is_pressed_in(KEY_RIGHT, keybuf_last),
			new_rg = is_pressed_in(KEY_RIGHT, keybuf_current);
		player->vx += (!old_rg && new_rg) - (old_rg && !new_rg);

		/* - check weapon status evolution - */
		int old_sh = is_pressed_in(KEY_SHIFT, keybuf_last),
			new_sh = is_pressed_in(KEY_SHIFT, keybuf_current);
		if (!old_sh && new_sh)
			player->weapon_status = wps_trig;
		if (old_sh && !new_sh)
			player->weapon_status = wps_hidden;

		/* - check if the player wants to exit - */
		if (!is_pressed_in(KEY_EXIT, keybuf_last)
			&& is_pressed_in(KEY_EXIT, keybuf_current)
			&& quit_popup())
			break;

		/* let ai decide what they want to do */
		ai_decide();

		/* sleep some time */
		sleep_some_time(1500);
	}

	/* If lost, display that the player lost */
	if (player->life <= 0) {
		char lost[] = "Ave Imperator.";

		int n = strlen(lost);
		int w = 2 + font_char_w * n + (n - 1) + 2;
		int h = 2 + font_char_h + 2;
		int x = DISPLAY_WIDTH / 2 - w / 2;
		int y = DISPLAY_HEIGHT / 2 - h / 2;

		dclear_area(x, y, x + w - 1, y + h - 1);
		dreverse_area(x, y, x + w - 1, y + h - 1);
		dreverse_area(x + 1, y + 1, x + w - 2, y + h - 2);
		dprint(x + 2, y + 2, lost);
		dupdate();
		while (1) {
			int key = getkey_opt(0, 0);
			if (key == KEY_EXE || key == KEY_EXIT) break;
		}
	}

	/* Free dynamically allocated tabs */
	free(ais);
	free(characters);

	return (last);
}
