#******************************************************************************#
#                                        _____           _                     #
#   tqmap/tqmap.py                      |_   _|__  _   _| |__   ___ _   _      #
#   | Project : TuQuoque/TQMap            | |/ _ \| | | | '_ \ / _ \ | | |     #
#                                         | | (_) | |_| | | | |  __/ |_| |     #
#   By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr  #
#   Last updated: 2016/08/10 21:10:10                               |___/      #
#                                                                              #
#******************************************************************************#
# Imports
from sys import stderr as err
import itertools as it

# Error messages
errmsg = {
	'width': "Warning: width is greater than 255, will be reduced\n",
	'height': "Warning: height is greater than 255, will be reduced\n",
	'sp_dup': "Warning: duplicate spawnpoint at ({}, {}), ignored\n",
	'sp_no': "Warning: no spawn point found, assuming ({}, {})\n",
	'esp_no': "Warning: no enemies spawn points found, is this a test map ?\n",
	'esp_max': "Warning: too many enemies spawn were found, only the 256 \
first ones will be kept\n",
	'no_support': "Error: {} at ({}, {}) with pattern {} isn't managed\n",
	'no_support_diags': "Error: {} at ({}, {}) with pattern {} and diagonal \
pattern {} isn't managed\n",
	'no_support_full': "Error: {} at ({}, {}) is surrounded by solids, \
unsupported\n",
	'no_support_corner': "Error: {} at ({}, {}) is at corner, unsupported\n",
	'no_support_opp': "Error: {} at ({}, {}) is between opposite walls, \
unsupported\n",
	'no_support_one': "Error: {} at ({}, {}) has one or less neighbours, \
unsupported\n",
	'unknown': "Error: character at ({}, {}) is unknown\n"
}

# Ennemies subclass
class TQMapEnemies:
	# Initialize
	def __init__(self, tab):
		self.data = tab

	# Make bytes
	def __bytes__(self):
		tab = [len(self.data), 0] + list(it.chain.from_iterable(self.data))
		return bytes(tab)

# Main class
class TQMap:
	# Initialize
	def __init__(self, air_char, player_spawn_char, enemy_spawn_char):
		# - Constants
		self.__air_char = air_char
		self.__sp_char = player_spawn_char
		self.__enemy_char = enemy_spawn_char
		# - Initializations
		self.__tiles = []
		self.__lines = []

	# Load a tile
	def add_tile(self, name, char, corresp, diags={}):
		# check args (never trust the user :D)
		if not type(name) is str:
			raise TypeError("name should be a string")
		if not type(char) is str or len(char) != 1:
			raise TypeError("char should be a character")
		''' --- As I'm the only user for now, that's okay, but think about
		    --- correcting this one day or an other

		if not corresp is dict and all([type(i) is tuple and \
			all([k in [0, 1] for k in i]) and j in range(255) \
			for i,j in corresp.items()]):
			raise TypeError("corresp hasn't got a correct format")
		if not diags is dict and all([type(i) is tuple and \
			all([k in [0, 1] for k in i]) and j in range(255) \
			for i,j in diags.items()]):
			raise TypeError("diags hasn't got a correct format")
		'''
		# then load
		self.__tiles.append((char, name, corresp, diags))

	# Load from file
	def load(self, file):
		# Get lines
		self.__lines = [l[:-1] if l[-1] == '\n' else l for l in file.readlines()]
		file.close()

	# Utilities
	# - get block
	def __gb(self, x, y):
		if x < 0 or x >= self.w or y < 0 or y >= self.h:
			return 1
		return int(self.__tab[y][x] != self.__air_char)
	# - get "type" (straight around blocks)
	def __gt(self, x, y):
		return (self.__gb(x, y - 1), self.__gb(x + 1, y), \
			self.__gb(x, y + 1), self.__gb(x - 1, y))
	# - get diagonal "type" (diagonal around blocks)
	def __gdt(self, x, y):
		return (self.__gb(x + 1, y - 1), self.__gb(x + 1, y + 1), \
			self.__gb(x - 1, y + 1), self.__gb(x - 1, y - 1))
	# - type is full
	def __t_isfull(self, t):
		return t == (1, 1, 1, 1)
	# - type is opposite
	def __t_isopp(self, t):
		return t in ((1, 0, 1, 0), (0, 1, 0, 1))
	# - type is in corner
	def __t_iscor(self, t):
		return t in ((1, 1, 0, 0), (0, 1, 1, 0), (0, 0, 1, 1), (1, 0, 0, 1))
	# - type is alone or has one neighbour
	def __t_aloneorone(self, t):
		return sum(t) <= 1

	# Process
	def process(self):
		# Get dimensions
		self.h, self.w = len(self.__lines), max(map(len, self.__lines))
		if self.w > 255:
			err.write(errmsg['width'])
		if self.h > 255:
			err.write(errmsg['height'])
		self.w, self.h = self.w % 256, self.h % 256

		# Make tab
		self.__tab = [[self.__lines[y][x] if x < len(self.__lines[y]) \
			else self.__air_char \
			for x in range(self.w)] for y in range(self.h)]
		del self.__lines

		# Look for player spawn
		self.spawn_x, self.spawn_y = -1, -1
		for y, x in it.product(range(self.h), range(self.w)):
			if self.__tab[y][x] == self.__sp_char:
				self.__tab[y][x] = self.__air_char
				if self.spawn_x == -1:
					self.spawn_x, self.spawn_y = x, y
				else:
					err.write(errmsg['sp_dup'].format(x, y))

		if self.spawn_x == -1:
			err.write(errmsg['sp_no'].format(1, 1))
			self.spawn_x, self.spawn_y = 1, 1

		# Look for ennemies spawn
		enemies = []
		for y, x in it.product(range(self.h), range(self.w)):
			if self.__tab[y][x] == self.__enemy_char:
				self.__tab[y][x] = self.__air_char
				enemies.append((x, y))
		count = len(enemies)
		if count == 0:
			err.write(errmsg['esp_no'])
		elif count > 256:
			err.write(errmsg['esp_max'])
			count %= 256
		self.enemies = TQMapEnemies(enemies[:count])

		# Get all blocks
		self.__final = [row[:] for row in self.__tab]
		waserr = 0
		for y, x in it.product(range(self.h), range(self.w)):
			# get "type" and diagonal "type"
			t, dt = self.__gt(x, y), self.__gdt(x, y)
			for char, name, cor, diags in self.__tiles:
				# check character
				if self.__tab[y][x] == char:
					if diags.get(t):
						if diags[t].get(dt):
							self.__final[y][x] = diags[t][dt]
						else:
							waserr = 1
							self.__final[y][x] = 0
							err.write(errmsg['no_support_diags'].format(name, x, y, t, dt))
					elif cor.get(t):
						self.__final[y][x] = cor[t]
					else:
						waserr = 1
						self.__final[y][x] = 0
						if self.__t_isfull(t):
							err.write(errmsg['no_support_full'].format(name, x, y))
						elif self.__t_isopp(t):
							err.write(errmsg['no_support_opp'].format(name, x, y))
						elif self.__t_iscor(t):
							err.write(errmsg['no_support_cor'].format(name, x, y))
						elif self.__t_aloneorone(t):
							err.write(errmsg['no_support_one'].format(name, x, y))
						else:
							err.write(errmsg['no_support'].format(name, x, y, t))
					continue
			if type(self.__final[y][x]) is str:
				if self.__tab[y][x] == self.__air_char:
					self.__final[y][x] = 0
				else:
					waserr = 1
					err.write(errmsg['unknown'].format(x, y))

		# Then return if there was an error
		del self.__tab
		return not waserr

	# - Generate bytestream
	def __bytes__(self):
		return bytes([self.w, self.h, self.spawn_x, self.spawn_y] \
			+ [self.__final[y][x] for y, x in it.product(range(self.h), range(self.w))])
